use uuid::Uuid;

use crate::constants::SSH_FILEXFER_ATTR_PERMISSIONS;
use crate::constants::SSH_FXP_ATTRS;
use crate::constants::SSH_FXP_HANDLE;
use crate::constants::SSH_FXP_INIT;
use crate::constants::SSH_FXP_LSTAT;
use crate::constants::SSH_FXP_NAME;
use crate::constants::SSH_FXP_OPENDIR;
use crate::constants::SSH_FXP_REALPATH;
use crate::constants::SSH_FXP_STAT;
use crate::constants::SSH_FXP_STATUS;
use crate::constants::SSH_FXP_VERSION;
use crate::constants::SSH_FX_FAILURE;
use crate::constants::SSH_FX_NO_SUCH_FILE;
use crate::constants::SSH_FX_OP_UNSUPPORTED;
use crate::constants::S_IFDIR;
use crate::constants::S_IFREG;
use crate::constants::S_IRWXU;

use crate::handle::DirectoryHandle;
use crate::packet::data::attrs::AttrsData;
use crate::packet::data::handle::HandleData;
use crate::packet::data::name::NameData;
use crate::packet::data::opendir::OpendirData;
use crate::packet::data::realpath::RealpathData;
use crate::packet::data::stat::StatData;
use crate::packet::data::status::StatusData;
use crate::packet::data::version::VersionData;
use crate::packet::header::PacketHeader;
use crate::packet::Packet;
use crate::php;
use crate::sftp::realpath;

use std::collections::HashMap;
use std::io::Read;
use std::io::{BufWriter, Write};

pub fn loop_of_dispatch(nextcloud_folder: &str, user: &str) {
    let mut directory_handles: HashMap<String, DirectoryHandle> = HashMap::new();
    loop {
        let packet = Packet::read_packet();
        dispatch_packet(packet, &mut directory_handles, nextcloud_folder, user)
    }
}

pub fn dispatch_packet(
    packet: Packet,
    directory_handles: &mut HashMap<String, DirectoryHandle>,
    nextcloud_folder: &str,
    user: &str,
) {
    let mut buff_stdout = BufWriter::new(std::io::stdout());
    let mut response: Packet = match packet.packet_header.type_packet {
        SSH_FXP_INIT => dispatch_version_request(packet),
        SSH_FXP_REALPATH => dispatch_realpath_request(packet),
        SSH_FXP_OPENDIR => {
            dispatch_opendir_request(packet, directory_handles, nextcloud_folder, user)
        }
        SSH_FXP_STAT => dispatch_stat_request(packet, nextcloud_folder, user),
        SSH_FXP_LSTAT => dispatch_stat_request(packet, nextcloud_folder, user),
        _ => {
            eprintln!("{} not implemented.", packet.packet_header.type_packet);
            let mut id: [u8; 4] = [0; 4];
            let mut packet_data: &[u8] = &packet.data;
            packet_data
                .read(&mut id)
                .expect("Id for the last request was not provided.");
            let id: u32 = u32::from_be_bytes(id);
            Packet {
                packet_header: PacketHeader {
                    length: 0,
                    type_packet: SSH_FXP_STATUS,
                },
                data: StatusData {
                    id: id,
                    status_code: SSH_FX_OP_UNSUPPORTED,
                    error_message: "The operation {} is not supported yet.".to_string(),
                    lang: "en_US".to_string(),
                }
                .serialize(),
            }
        }
    };
    let serialized_response = response.serialize();
    eprintln!("Writing response.");
    eprintln!("{:#?}", &serialized_response);
    buff_stdout
        .write(&serialized_response)
        .expect("Unable to write to stdout.");
    buff_stdout.flush().expect("Unable to flush stdout.");
}

fn dispatch_stat_request(packet: Packet, nextcloud_folder: &str, user: &str) -> Packet {
    let stat_data: StatData = StatData::deserialize(&packet.data);
    if !php::check_if_exists(&stat_data.path, &user, &nextcloud_folder) {
        Packet {
            packet_header: PacketHeader {
                length: 0,
                type_packet: SSH_FXP_STATUS,
            },
            data: StatusData {
                id: stat_data.id,
                status_code: SSH_FX_NO_SUCH_FILE,
                error_message: "No such file or directory.".to_string(),
                lang: "en_US".to_string(),
            }
            .serialize(),
        }
    } else {
        let mut permission: u32 = 0;
        let mut attrs: u32 = 0;
        attrs |= SSH_FILEXFER_ATTR_PERMISSIONS;
        permission |= S_IRWXU;
        if php::check_if_is_dir(&stat_data.path, &user, &nextcloud_folder) {
            permission |= S_IFDIR;
        } else {
            permission |= S_IFREG;
        }
        Packet {
            packet_header: PacketHeader {
                length: 0,
                type_packet: SSH_FXP_ATTRS,
            },
            data: AttrsData::new(stat_data.id, attrs, permission).serialize(),
        }
    }
}

fn dispatch_opendir_request(
    packet: Packet,
    directory_handles: &mut HashMap<String, DirectoryHandle>,
    nextcloud_folder: &str,
    user: &str,
) -> Packet {
    let opendir_packet_data = OpendirData::deserialize(&packet.data);
    let path = opendir_packet_data.path;
    if !php::check_if_exists(&path, &user, &nextcloud_folder) {
        return Packet {
            packet_header: PacketHeader {
                length: 0,
                type_packet: SSH_FXP_STATUS,
            },
            data: StatusData {
                id: opendir_packet_data.id,
                status_code: SSH_FX_NO_SUCH_FILE,
                error_message: "No such file or directory.".to_string(),
                lang: "en_US".to_string(),
            }
            .serialize(),
        };
    }
    if !php::check_if_is_dir(&path, &user, &nextcloud_folder) {
        return Packet {
            packet_header: PacketHeader {
                length: 0,
                type_packet: SSH_FXP_STATUS,
            },
            data: StatusData {
                id: opendir_packet_data.id,
                status_code: SSH_FX_FAILURE,
                error_message: "The file is not a directory.".to_string(),
                lang: "en_US".to_string(),
            }
            .serialize(),
        };
    }
    let str_handle: String = Uuid::new_v4().to_string();
    eprintln!("{}", str_handle);
    directory_handles.insert(
        str_handle.clone(),
        DirectoryHandle::new(php::ls(&path, &user, &nextcloud_folder)),
    );
    Packet {
        packet_header: PacketHeader {
            length: 0,
            type_packet: SSH_FXP_HANDLE,
        },
        data: HandleData {
            id: opendir_packet_data.id,
            handle: str_handle,
        }
        .serialize(),
    }
}

fn dispatch_realpath_request(packet: Packet) -> Packet {
    let realpath_packet_data: RealpathData = RealpathData::deserialize(&packet.data);
    let path = realpath(&realpath_packet_data.path);
    let id = realpath_packet_data.id;
    Packet {
        packet_header: PacketHeader {
            length: 0,
            type_packet: SSH_FXP_NAME,
        },
        data: NameData::new(id, &[path]).serialize(),
    }
}

fn dispatch_version_request(packet: Packet) -> Packet {
    let init_packet_data: VersionData = match VersionData::deserialize(&packet.data) {
        Ok(ok) => ok,
        Err(err) => panic!(err),
    };
    eprintln!("The client version is {}.", init_packet_data.version);
    if init_packet_data.version < 3 {
        panic!("Unsupported version, minimal client version 3.");
    }
    let version_packet_data = VersionData {
        version: 3,
        extension_data: Vec::new(),
    };
    Packet {
        packet_header: PacketHeader {
            length: 0,
            type_packet: SSH_FXP_VERSION,
        },
        data: version_packet_data.serialize(),
    }
}
